#include <algorithm>
#include <iostream>
#include <cstdlib>
#include <chrono>
#include <omp.h>

int N;          // data array size
int *a;         // data array to be sorted

void fillArray(void);
static inline void exchange(int i, int j);
void bitonicSort(void);

int main(int argc, char **argv) {
	if (argc != 2) {
		std::cout << "Usage: " << argv[0] << " n - where 2^n is problem size" << std::endl;
		exit(1);
	}

	int power = atoi(argv[1]);

	if (power < 1 || power>30) {
		std::cout << "n must be bigger than 0 and smaller than 30" << std::endl;
		exit(1);
	}

	N = 1 << power;
	a = new int[N];

	fillArray();

	auto bitonic_start = std::chrono::high_resolution_clock::now();
	bitonicSort();
	auto bitonic_finish = std::chrono::high_resolution_clock::now();

	std::chrono::duration<double> bitonic_elapsed = bitonic_finish - bitonic_start;

	if (std::is_sorted(a, a + N)) {
		std::cout << "Bitonic sort time: " << bitonic_elapsed.count() << std::endl;
	}
	else {
		std::cout << "Bitonic sort test failed!" << std::endl;
	}

	fillArray();
	auto std_start = std::chrono::high_resolution_clock::now();
	std::sort(a, a + N);
	auto std_finish = std::chrono::high_resolution_clock::now();

	std::chrono::duration<double> std_elapsed = std_finish - std_start;

	if (std::is_sorted(a, a + N)) {
		std::cout << "Std sort time: " << std_elapsed.count() << std::endl;
	}
	else {
		std::cout << "Std sort test failed!" << std::endl;
	}

	delete[] a;
}

void fillArray() {
	for (int i = 0; i < N; i++) {
		// a[i] = rand() % N; // (N - i);
		a[i] = (N - i);
	}
}

static inline void exchange(int i, int j) {
	int t = a[i];
	a[i] = a[j];
	a[j] = t;
}

void bitonicSort() {
	for (int k = 2; k <= N; k = 2 * k) {
		for (int j = k >> 1; j > 0; j = j >> 1) {
			for (int i = 0; i < N; i++) {
				int ij = i^j;
				if ((ij) > i) {
					if ((i&k) == 0 && a[i] > a[ij])
						exchange(i, ij);
					if ((i&k) != 0 && a[i] < a[ij])
						exchange(i, ij);
				}
			}
		}
	}
}
